/*

  Dismantle Esterel statements into high-level SUIF statements
  (while, if, etc.)

 */

#include "statement_dismantlers.h"
#include <basicnodes/basic_factory.h>
#include <suifnodes/suif_factory.h>
#include <esuifnodes/esuif_factory.h>
#include <utils/expression_utils.h>
#include <utils/symbol_utils.h>
#include <utils/type_utils.h>
#include <suifkernel/utilities.h>
#include <common/lstring.h>
#include <typebuilder/type_builder.h>

// FIXME: only for debugging
#include <stdio.h>

static LString lessThanOp("is_less_than");
static LString subtractOp("subtract");
static LString levelVarAnnote("level_var");
QualifiedType *ParallelTrapExitModule::_integer_type = 0;

static LString inputDirection("input");
static LString inputoutputDirection("inputoutput");
static LString outputDirection("output");

static LString tickSignal("tick");

// Helper function for statement lists

inline StatementList &operator<<(StatementList &sl, Statement *s)
{
  sl.append_statement(s);
  return sl;
}

/**
   Create a new variable by adding a unique numeric suffix to
   the given base name.
 */
static VariableSymbol *new_variable(SuifEnv *env, SymbolTable *st,
				    QualifiedType *qt, String base)
{
  String name;
  for (IInteger i = 0 ; ; i++) {
    name = base + i.to_String();
    if (!LString::exists(name)) {
      VariableSymbol *result = create_variable_symbol(env, qt);
      st->add_symbol(name, result);
      return result;
    }
  }
}

/**
   Create a new label by adding a unique numeric suffix to the given
   base name.   
 */
static CodeLabelSymbol *new_label(SymbolTable *st, String base)
{
  SuifEnv *env = st->get_suif_env();
  String name;
  for (IInteger i = 0 ; ; i++) {
    name = base + i.to_String();
    if (!LString::exists(name)) {
      CodeLabelSymbol *result =
	create_code_label_symbol(env, retrieve_label_type(env));
      st->add_symbol(name, result);
      return result;
    }
  } 
}

/**
  Remove annotations from the first object and attach them to the second
 */
void move_annotes(AnnotableObject *src, AnnotableObject *dest)
{
  while (src->get_annote_count()) {
    Annote *a = src->get_annote(0);
    src->remove_annote(0);
    dest->append_annote(a);
  }
}

/**
  Create a resume statement and its variables  
*/

ResumeStatement *create_resume_with_vars(SuifEnv *env, ProcedureDefinition *pd,
					 Statement *body,
					 bool shouldInitialize)
{
  CodeLabelSymbol *exit_label = new_label(pd->get_symbol_table(), "Exit");

  EsterelInformationBlock *info_block =
    to<EsterelInformationBlock>(get_global_information_block(env,
		   EsterelInformationBlock::get_class_name()));
  suif_assert(info_block); // Should have been attached by strl2suif

  TypeBuilder *tb = get_type_builder(env);
  suif_assert(tb);

  QualifiedType *integer_type =
    tb->get_qualified_type(info_block->get_integer_type());

  // Create the static state variable in the file's scope

  suif_assert(pd->get_parent());
  FileBlock *fb = to<FileBlock>(pd->get_parent()->get_parent());

  VariableSymbol *state_variable =
    new_variable(env, fb->get_symbol_table(), integer_type, "_state_");

  // Initialize the static variable to 0 if the flag is given
  ValueBlock *vb =
    shouldInitialize ?
    create_expression_value_block(env, create_int_constant(env, info_block->get_integer_type(), 0))
    : 0;

  VariableDefinition *vd =
    create_variable_definition(env, state_variable, 0, vb, true);
  fb->get_definition_block()->append_variable_definition(vd);

  return create_resume_statement(env, body, exit_label, state_variable);
}

/**
  Create a statement of the form <int-variable> := k
 */
StoreVariableStatement *create_set_variable_constant(SuifEnv *env,
						     VariableSymbol *vs,
						     int constant)
{
  IntConstant *ic = create_int_constant(env, vs->get_type()->get_base_type(),
					constant);
  return create_store_variable_statement(env, vs, ic);
}

/**
   Create an expression returning the value of a variable.  Its type
   is taken from the variable symbol.
*/
LoadVariableExpression *create_load_variable( SuifEnv *env,
					      VariableSymbol *vs)
{
  return create_load_variable_expression( env,
					  vs->get_type()->get_base_type(),
					  vs );
}

/**
  Dismantle an Esterel CaseStatement (present, if, await)
  into a cascade of IfStatements.

  present
    case p1 do s1
    case p2 do s2
    else s3
  end

  if (p1) s1
  else if (p2) s2
       else s3
*/
static IfStatement *dismantle_case_statement(SuifEnv *env, CaseStatement *cs)
{
  IfStatement *result = 0;
  IfStatement *lastif = 0;

  suif_assert(cs->get_case_count() > 0); // Need at least one case

  for ( Iter<PredicatedStatement*> iter = cs->get_case_iterator() ;
	iter.is_valid() ; iter.next() ) {
    PredicatedStatement *prs = iter.current();
    Expression *e = prs->get_predicate();
    suif_assert(e);
    prs->set_predicate(0);
    // remove_suif_object(e);
    Statement *s = prs->get_body();
    // if (s) remove_suif_object(s);
    prs->set_body(0);
    IfStatement *thisif = create_if_statement( env, e, s );
    if (result) {
      lastif->set_else_part(thisif); // This is a later case: make it an "else" branch
    } else {
      result = thisif; // This is the first case: make it the result
    }
    lastif = thisif;
  }

  suif_assert(result); // Should have created at least one

  // Put the default case, if any, into the "else" part of the last IfStatement

  Statement *s = cs->get_default_case();
  if (s) {
    cs->set_default_case(0);
    // remove_suif_object(s);
    lastif->set_else_part(s);
  }

  return result;
}

AnnotableObject *Dismantler<PresentStatement>::dismantle(PresentStatement *s)
{
  return dismantle_case_statement(env, s);
}

AnnotableObject *
Dismantler<EsterelIfStatement>::dismantle(EsterelIfStatement *s)
{
  return dismantle_case_statement(env, s);
}

/**
   await
    case immediate p1 do s1
    case p2 do s2
    case p3 do s3
   end

   abort
    loop pause end
   when
    case immediate p1 do s1
    case p2 do s2
    case p3 do s3
   end     
 */
AnnotableObject * Dismantler<AwaitStatement>::dismantle(AwaitStatement *as)
{
  PauseStatement *pause = create_pause_statement(env);
  LoopStatement *body = create_loop_statement(env, pause);

  AbortStatement *result = create_abort_statement(env, body);

  // Copy the predicates from the Await statement to the newly-created abort

  for ( size_t i = 0 ; i < as->get_case_count() ; ++i ) {
    PredicatedStatement *ps = as->get_case(i);
    as->replace_case(i,0);
    result->append_case(ps);
  }

  return result;
}

/**
    abort
      body
    when
      case immediate p1 do s1
      case p2 do s2
      case count p3 do s3
    end
    

    % abort

    if p1 goto S1;
    c = count;
    goto Entry
    Restart:
      if p1 then S1: s1
      else if p2 then s2
      else if p3 && !(--c >= 0) then s2
      else
        resume {
          Entry:
            body
            goto Exit
        }
        pause
        goto Restart
    Exit:

    % weak abort case

    if p1 goto S1;
    c = count;
    goto Entry
    Restart:
    resume {
            Entry:
            body
            goto Exit
           }
    if p1 then S1: s1
    else if p2 then s2
    else if p3 && !(--c >= 0) then s2
    else
      pause
      goto Restart
    Exit:
 */
AnnotableObject *Dismantler<AbortStatement>::dismantle(AbortStatement *as)
{
  StatementList *result = 0;

  ProcedureDefinition *proc_def = get_procedure_definition(as);
  SymbolTable *st = proc_def->get_symbol_table();

  Statement *body = as->get_body();
  as->set_body(0);
  suif_assert(body);

  CodeLabelSymbol *restart_label = new_label(st, "Restart");
  CodeLabelSymbol *entry_label = new_label(st, "Entry");
  CodeLabelSymbol *exit_label = new_label(st, "Exit");

  StatementList *resume_body = create_statement_list(env);

  *resume_body << create_label_location_statement(env, entry_label)
	       << body
	       << create_jump_statement(env, exit_label);

  suif_assert(as->get_case_count() > 0); // Should be at least one case

  StatementList *initialize = create_statement_list(env);
  IfStatement *firstif = 0;
  IfStatement *lastif = 0;

  for (Iter<PredicatedStatement*> iter = as->get_case_iterator() ;
       iter.is_valid() ; iter.next() ) {
    PredicatedStatement *ps = iter.current();
    Expression *e = ps->get_predicate();
    suif_assert(e);
    ps->set_predicate(0);

    Statement *handler = ps->get_body();
    ps->set_body(0);

    if ( is_a<Delay>(e) ) {
      Delay *d = to<Delay>(e);

      if (d->get_is_immediate()) {
	StatementList *newhandler = create_statement_list(env);
	CodeLabelSymbol *handler_label = new_label(st, "Handler");
	*newhandler << create_label_location_statement(env, handler_label);
	e = d->get_predicate();
	d->set_predicate(0);
	*initialize <<
	  create_if_statement(env, e,
			      create_jump_statement(env, handler_label));
	e = to<Expression>(e->deep_clone()); // Expression will be used again
	if (handler) *newhandler << handler;
	handler = newhandler;
      }

      // FIXME: Handle counted delays
      // Create a new count variable
      // Initialize it in the initialize portion
      // test and decrement in the expression

    }

    IfStatement *thisif = create_if_statement(env, e, handler);
    if (lastif) {
      lastif->set_else_part(thisif);
    } else {
      firstif = thisif;
    }
    lastif = thisif;
  }

  suif_assert(firstif);

  result = create_statement_list(env);
  *result << initialize
	  << create_jump_statement(env, entry_label)
	  << create_label_location_statement(env, restart_label);

  StatementList *elsebody = create_statement_list(env);  

  ResumeStatement *resume =
    create_resume_with_vars(env, proc_def, resume_body, false);

  if (as->get_weak()) {
    *result << resume; // Weak case: put the body before the tests
  } else {
    *elsebody << resume; // Strong case: put the body after the tests
  }
  
  *elsebody << create_pause_statement(env)
	    << create_jump_statement(env, restart_label);

  lastif->set_else_part(elsebody);
  *result << firstif
	  << create_label_location_statement(env, exit_label);

  suif_assert(result);
  return result;
}

/**
   do b watching p timeout s

   abort
     b
   when p do s
*/
AnnotableObject *
Dismantler<DoWatchingStatement>::dismantle(DoWatchingStatement *dws)
{
  Statement *body = dws->get_body();
  dws->set_body(0);
  suif_assert(body);

  AbortStatement *result = create_abort_statement(env, body);

  Statement *timeout = dws->get_timeout();
  dws->set_timeout(0);

  Expression *predicate = dws->get_predicate();
  dws->set_predicate(0);
  suif_assert(predicate);

  PredicatedStatement *ps =
    create_predicated_statement(env, timeout, predicate );
  result->append_case(ps);

  return result;
}

/**
   do b upto p timeout s

   abort
     b; halt
   when p do s
*/
AnnotableObject *Dismantler<DoUptoStatement>::dismantle(DoUptoStatement *dus)
{
  Statement *body = dus->get_body();
  dus->set_body(0);
  suif_assert(body);

  StatementList *abort_body = create_statement_list(env);
  *abort_body << body
	      << create_halt_statement(env);

  AbortStatement *result = create_abort_statement(env, abort_body);

  Expression *predicate = dus->get_predicate();
  dus->set_predicate(0);
  suif_assert(predicate);

  PredicatedStatement *ps = create_predicated_statement( env, 0, predicate );
  result->append_case(ps);

  return result;
}

/**
   loop b each p

   loop
     abort
       b; halt
     when p
   end
*/
AnnotableObject *
Dismantler<LoopEachStatement>::dismantle(LoopEachStatement *les)
{
  Statement *body = les->get_body();
  les->set_body(0);
  suif_assert(body);

  StatementList *abort_body = create_statement_list(env);
  *abort_body << body
	      << create_halt_statement(env);

  AbortStatement *abort =
    create_abort_statement(env, abort_body);

  Expression *predicate = les->get_predicate();
  les->set_predicate(0);
  suif_assert(predicate);

  PredicatedStatement *ps = create_predicated_statement( env, 0, predicate );
  abort->append_case(ps);

  LoopStatement *result = create_loop_statement( env, abort );

  return result;
}

/**
  halt

  loop
    pause
  end
 */
AnnotableObject *Dismantler<HaltStatement>::dismantle(HaltStatement *hs)
{  
  PauseStatement *ps = create_pause_statement(env);

  LoopStatement *result = create_loop_statement(env, ps);
  return result;
}

/**
  sustain s

  loop
    emit s;
    pause
  end
 */
AnnotableObject *
Dismantler<SustainStatement>::dismantle(SustainStatement *ss)
{
  SignalSymbol *signal = ss->get_destination();
  Expression *e = ss->get_value();
  ss->set_value(0);

  StatementList *sl = create_statement_list(env);
  *sl << create_emit_statement(env, signal, e)
      << create_pause_statement(env);

  LoopStatement *result = create_loop_statement(env, sl);
  
  return result;
}

/**
  loop b end

  Restart:
    b
    goto Restart
 */
AnnotableObject *Dismantler<LoopStatement>::dismantle(LoopStatement *ls)
{
  Statement *body = ls->get_body();
  ls->set_body(0);
  suif_assert(body);
  
  StatementList *result = create_statement_list(env);

  CodeLabelSymbol *restart_label =
    new_label(get_procedure_definition(ls)->get_symbol_table(), "Loop");

  *result << create_label_location_statement(env, restart_label)
	  << body
	  << create_jump_statement(env, restart_label);
  
  return result;
}

/**
  every p do b
  
  loop
    abort
      b;
    Entry: halt
    when p
  end

  This is not entirely accurate: the abort statement is special: the entry
  point is later.
 */
// FIXME: dismantle_every_do_statement
// FIXME: Use part of the abort machinery to do this expansion

/**

  Dismantle parallel, trap, and exit statements

 */

void ParallelTrapExitModule::do_procedure_definition(ProcedureDefinition *pd)
{
  if (is_a<ModuleDefinition>(pd)) {
    SuifEnv *env = pd->get_suif_env();
    // Locate the built-in "integer" type for use with exit levels

    TypeBuilder *tb = get_type_builder(env);

    EsterelInformationBlock *info_block =
      to<EsterelInformationBlock>(get_global_information_block(env,
							       EsterelInformationBlock::get_class_name()));
    suif_assert(info_block); // Should have been attached by strl2suif

    _integer_type = tb->get_qualified_type(info_block->get_integer_type());

    ModuleDefinition *md = to<ModuleDefinition>(pd);  
    SymbolTable *st = md->get_symbol_table();

    // Surround the body of the procedure with a thread containing a resume

    Statement *body = to<Statement>(pd->get_body());
    pd->set_body(0); // Relinquish ownership

    StatementList *resume_body = create_statement_list(env);
    *resume_body << create_pause_statement(env)
		 << body
		 << create_exit_at_statement(env, 0);

    StatementList *thread_body = create_statement_list(env);
    *thread_body << create_resume_with_vars(env, md, resume_body, true)
		 << create_exit_at_statement(env, 1);

    ThreadStatement *ts = create_thread_statement(env, thread_body);
    ts->set_exit_var(new_variable(env, st, _integer_type, "_exit_"));
    ts->set_state_var(new_variable(env, st, _integer_type, "_state_"));
    ts->set_join_label(new_label(st, "Join"));
					       
    pd->set_body(ts);

    // Dismantle TrapStatements into TrapScopeStatements
    Dismantler<TrapStatement> dismantle_traps(env, pd);
    pd->walk(dismantle_traps);
  
    // Add join labels, state, and level variables to threads
    ObjectWalker<ParallelStatementList> add_thread_variables(env, pd);
    pd->walk(add_thread_variables);

    // Dismantle Exit statements into assignments and gotos
    Dismantler<ExitStatement> esw(env, pd);
    pd->walk(esw);

    // Dismantle parallel statements into fork/join and exit level variable tests
    Dismantler<ParallelStatementList> pslw(env,pd);
    pd->walk(pslw);
  }
}

/**
  Dismantle TrapStatements into TrapScopeStatement and handlers

  Replace the handlers with a single statement, either code
  for a single handler or a parallel statement list for multiple ones.
  Tests are only created if there is more than one trap symbol
  Parallel are created when there is more than one handler.
  This may create ParallelStatementLists, so dismantle TrapStatements first

  trap T1 := e : integer, T2 in
    body
  handle T1 do h1
  handle T2 do h2

  T1_presence = false
  T1_val = e
  TrapScopeStatement  T1, T2
    body
    goto ExitLabel
HandlerLabel:
  [
    if T1 {h1}
  ||
    if T2 {h2}
  ]
ExitLabel:

*/

AnnotableObject *Dismantler<TrapStatement>::dismantle(TrapStatement *ts)
{
  SymbolTable *st = get_procedure_definition(ts)->get_symbol_table();

  StatementList *result = create_statement_list(env);

  // Create assignments that reset the traps and initialize them if necessary

  for ( Iter<SymbolTableObject *> iter =
	  ts->get_symbol_table()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    if (is_a<TrapSymbol>(iter.current())) {
      TrapSymbol *ts = to<TrapSymbol>(iter.current());

      VariableSymbol *presence = ts->get_presence();
      *result << create_set_variable_constant(env, presence, 0);

      if (ts->get_value()) {
	VariableDefinition *vd = ts->get_value()->get_definition();
	suif_assert(vd);
	if (vd->get_initialization()) {
	  ExpressionValueBlock *evb =
	    to<ExpressionValueBlock>(vd->get_initialization());
	  Expression *e = evb->get_expression();
	  suif_assert(e);
	  evb->set_expression(0);
	  VariableSymbol *vs = ts->get_value();
	  suif_assert(vs); // Initializing expression should imply value var
	  *result << create_store_variable_statement(env, vs, e);
	}
      }
    }
  }

  // Create a TrapScopeStatement with the body, symbol table, and
  // definition block from the Trap statement

  Statement *trap_body = ts->get_body();
  ts->set_body(0);
  SymbolTable *trap_st = ts->get_symbol_table();
  ts->set_symbol_table(0);
  DefinitionBlock *trap_db = ts->get_definition_block();
  ts->set_definition_block(0);

  StatementList *new_body = create_statement_list(env);
  *new_body << trap_body;
  CodeLabelSymbol *exit_label = new_label(st, "Exit");
  *new_body << create_jump_statement(env, exit_label);

  TrapScopeStatement *tss =
    create_trap_scope_statement(env, new_label(st, "Trap"),
				new_body, trap_st, trap_db);
  *result << tss;

  // HandlerLabel:

  *result << create_label_location_statement(env, tss->get_handler_label());

  // Create handlers
  // May be nothing, a single handler, a conditional handler, or
  // handlers in parallel

  // Need to check the predicates if there's more than one trap
  bool need_tests = trap_st->get_symbol_table_object_count() > 1;

  // Need to run handlers in parallel if there's more than one
  bool in_parallel = ts->get_handler_count() > 1;

  ParallelStatementList *psl;
  if (in_parallel) {
    psl = create_parallel_statement_list(env);
    *result << psl;
  }

  while ( ts->get_handler_count() ) {
    PredicatedStatement *ps = to<PredicatedStatement>(ts->get_handler(0));
    ts->remove_handler(0);
    Expression *e = ps->get_predicate();
    ps->set_predicate(0);
    Statement *s = ps->get_body();
    ps->set_body(0);
    if (need_tests) s = create_if_statement(env, e, s);
    if (in_parallel) {
      psl->append_thread(create_thread_statement(env, s));
    } else {
      *result << s;
    }
  }

  // ExitLabel:

  *result << create_label_location_statement(env, exit_label);

  return result;
}

/**
  Add a join label and variables for exit levels and state to each thread
  of a parallel statement list

  This needs to be run before ParallelStatementLists are dismantled
*/
void ObjectWalker<ParallelStatementList>::walk(ParallelStatementList *psl)
{
  SymbolTable *st = get_procedure_definition(psl)->get_symbol_table();

  QualifiedType *qt = ParallelTrapExitModule::get_integer_type();

  CodeLabelSymbol *join_label = new_label(st, "Join");

  for ( Iter<ThreadStatement *> iter = psl->get_thread_iterator() ;
	iter.is_valid() ; iter.next() ) {
    ThreadStatement *ts = iter.current();
    ts->set_join_label(join_label);
    ts->set_exit_var(new_variable(env, st, qt, "_exit_"));
    ts->set_state_var(new_variable(env, st, qt, "_state_"));
  }
}

/**
  exit T(e)

  T_presence := 1
  T_value := e
  exit @k       // When within a parallel
  goto Handler
 
  Handler is either the handler for the trap statement itself, or the join
  label of the parallel statement this exit resides it

  This can only run before ThreadStatements are dismantled.
*/
AnnotableObject *Dismantler<ExitStatement>::dismantle(ExitStatement *es)
{
  // Look up the ownership tree for traps and parallels
  // to compute the exit level

  int exit_level = 2;
  ThreadStatement *enclosing_thread = 0;
  TrapScopeStatement *enclosing_trap = 0;
  SuifObject *so;
  for ( so = es ; so ; so = so->get_parent()) {
    if (is_a<TrapScopeStatement>(so)) {
      enclosing_trap = to<TrapScopeStatement>(so);
      if ( es->get_trap()->get_parent() ==
	   enclosing_trap->get_symbol_table() )
	break;
      if (enclosing_thread) exit_level++;
    } else if (!enclosing_thread && is_a<ThreadStatement>(so)) {
      enclosing_thread = to<ThreadStatement>(so);
    }
  }

  suif_assert_message(so,
	      ("did not find trap %s", es->get_trap()->get_name().c_str()));

  StatementList *result = create_statement_list(env);

  // Set the trap presence variable

  VariableSymbol *presence = es->get_trap()->get_presence();
  *result << create_set_variable_constant(env, presence, 1);

  // If there's an expression, store it in the value variable
  VariableSymbol *value = es->get_trap()->get_value();
  Expression *e = es->get_value();
  if (e) {
    suif_assert(value); // Should have a valued trap if there's an expression
    es->set_value(0);
    *result <<  create_store_variable_statement(env, value, e);
  }							   

  if (enclosing_thread) {

    // Trap found outside this thread: set the thread's exit level and
    // branch to the join for the thread

    *result << create_exit_at_statement(env, exit_level);

  } else {

    // Trap found within the thread: jump to its handler

    suif_assert(enclosing_trap);
    *result << create_jump_statement(env, enclosing_trap->get_handler_label());
  }

  suif_assert(result);
  return result;
}

/**
  b1 || b2

  fork E1 E2

Again:
  pause;
  fork R1 R2

  thread {
  R1:
    resume {
    E1:
      b1
      exit @0
    }
    exit @1
  }

  thread {
  R2:
    resume {
    E2:
      b2
      exit @0
    }
    exit @1
  }

Join:
  join
  level = max of thread's levels
  switch (exit level) {
  case 0: goto Exit
  case 1: goto Again
  case 2: goto Handler
  case 3: goto Handler
  case 4: goto Handler1
  case 5: goto Handler2
  }
Handler1:
  exit @2
Handler2:
  exit @3
Exit: 

This must be run before traps are dismantled since it looks for them
to compute exit levels.
*/

// FIXME: add thread exit level reset statements before the forks

AnnotableObject *
Dismantler<ParallelStatementList>::dismantle(ParallelStatementList *psl)
{
  StatementList *result = create_statement_list(env);

  // fork Entry1, Entry2, ...

  ForkStatement *entry_fork = create_fork_statement(env);
  *result << entry_fork;

  // Again:
  //   pause
  //   fork Resume1, Resume2, ...
  ProcedureDefinition *pd = get_procedure_definition(psl);
  SymbolTable *st = pd->get_symbol_table();
  CodeLabelSymbol *again_label = new_label(st, "Again");
  ForkStatement *resume_fork = create_fork_statement(env);
  *result << create_label_location_statement(env, again_label)
	  << create_pause_statement(env)
	  << resume_fork;

  // This collection should have been constructed with at least one thread
  suif_assert(psl->get_thread_count() > 0);

  StatementList *max_statement = 0;
  if ( psl->get_thread_count() > 1 )
    max_statement = create_statement_list(env);

  VariableSymbol *first_level = 0;

  CodeLabelSymbol *join_label = psl->get_thread(0)->get_join_label();

  while ( psl->get_thread_count() ) {
    ThreadStatement *thread = psl->get_thread(0);
    psl->remove_thread(0);

    VariableSymbol *vs = thread->get_exit_var();

    if (first_level) {
      suif_assert(max_statement); // Should know there was more than one

      // Create "if (level1 < levelk) { level1 = levelk; }"

      DataType *ty = first_level->get_type()->get_base_type();

      Expression *e =
	create_binary_expression(env, ty, lessThanOp,
				 create_load_variable(env, first_level),
				 create_load_variable(env, vs));
      *max_statement <<
        create_if_statement(env, e,
 	  create_store_variable_statement(env, first_level,
					  create_load_variable(env, vs)));
    } else {
      first_level = vs;
    }

    StatementList *body = create_statement_list(env);

    // Resume:
    CodeLabelSymbol *resume_label = new_label(st, "Resume");
    *body << create_label_location_statement(env, resume_label);
    resume_fork->append_target(resume_label);

    // Body of resume statement
    //  Entry:
    StatementList *resume_body = create_statement_list(env);
    CodeLabelSymbol *entry_label = new_label(st, "Entry");
    *resume_body << create_label_location_statement(env, entry_label);
    entry_fork->append_target(entry_label);
    //   body
    Statement *old_body = thread->get_body();
    thread->set_body(0);

    *resume_body << old_body
		 << create_exit_at_statement(env, 0); // exit @0

    *body << create_resume_with_vars(env, pd, resume_body, false)
	  << create_exit_at_statement(env, 1);

    thread->set_body(body);
    *result << thread;
  }

  *result << create_join_statement(env, join_label);

  // Add the statement(s), if any, that compute the maximum exit level
  if (max_statement)
    *result << max_statement;

  CodeLabelSymbol *exit_label = new_label(st, "Exit");

  suif_assert(first_level); // Should always find one level variable

  MultiWayBranchStatement * mwbs =
    create_multi_way_branch_statement(env,
     create_load_variable_expression(env, first_level->get_type()->get_base_type(),
				     first_level), 0);
  *result << mwbs;

  mwbs->add_case(0, exit_label);
  mwbs->add_case(1, again_label);

  // Handle cases 2...k by branching to their trap handlers
  int level_handled = 2;
  SuifObject *so;
  for ( so = psl->get_parent() ; !is_a<ThreadStatement>(so) ;
	so = so->get_parent() ) {
    if (is_a<TrapScopeStatement>(so)) {
      TrapScopeStatement *ts = to<TrapScopeStatement>(so);
      mwbs->add_case(level_handled, ts->get_handler_label());
      level_handled++;
    }
  }

  // Handle cases k...n by exiting the thread at a reduced level
  int level_offset = level_handled - 2;
  for ( ; so ; so = so->get_parent() ) {
    if (is_a<TrapScopeStatement>(so)) {
      CodeLabelSymbol *handler_label = new_label(st, "Handler");
      *result << create_label_location_statement(env, handler_label)
	      << create_exit_at_statement(env, level_handled - level_offset);
      mwbs->add_case(level_handled, handler_label);
      level_handled++;
    }
  }
 
  *result << create_label_location_statement(env, exit_label);
 
  return result;
}

/**
  emit S(e)

  S_presence = true
  S_value = e
 */
AnnotableObject *
Dismantler<EmitStatement>::dismantle(EmitStatement *es)
{
  SignalSymbol *ss = es->get_destination();
  suif_assert(ss);

  StatementList *result = create_statement_list(env);

  VariableSymbol *presence = ss->get_presence();
  suif_assert(presence);
  *result << create_set_variable_constant(env, presence, 1);

  VariableSymbol *value = ss->get_value();
  Expression *e = es->get_value();
  if (e) {
    suif_assert(value); // Expression means this is a valued signal
    es->set_value(0);
    *result << create_store_variable_statement(env, value, e);
  }

  return result;
}

/**
   Dismantle a LoadTrapExpression into a load variable expression for the
   trap's presence variable.

 */
AnnotableObject *
Dismantler<LoadTrapExpression>::dismantle(LoadTrapExpression *lte)
{
  TrapSymbol *ts = lte->get_source();
  suif_assert(ts);
  DataType *t = ts->get_presence()->get_type()->get_base_type();
  return create_load_variable_expression(env, t, ts->get_presence());
}

/**

  Dismantle a ThreadStatement into its body (i.e., remove the scope)

  This destroys a lot of structure and should be done very late in the process.

 */
AnnotableObject *
Dismantler<ThreadStatement>::dismantle(ThreadStatement *ts)
{
  Statement *result = ts->get_body();
  ts->set_body(0);
  return result;
}

/**
  pause;

  resume_var := k;
  goto ResumeExit;
case k:

 */

AnnotableObject *
Dismantler<PauseStatement>::dismantle(PauseStatement *ps)
{
  StatementList *result = create_statement_list(env);

  // Locate the resume statement that encloses this pause

  SuifObject *so;
  for ( so = ps ; so ; so = so->get_parent() ) {
    if (is_a<ResumeStatement>(so)) break;
  }
  suif_assert_message(so, ("resume statement for pause not found"));

  ResumeStatement *rs = to<ResumeStatement>(so);

  // Encoding of each pause statement is trivial: the first one attached
  // to a resume statement is assigned 0, the next 1, etc.

  int pause_encoding = rs->get_resume_point_count();

  VariableSymbol *state_var = rs->get_state_var();
  suif_assert(state_var);

  SymbolTable *st = get_procedure_definition(ps)->get_symbol_table();
  CodeLabelSymbol *resume_label = new_label(st, "Resume");
  rs->append_resume_point(resume_label);

  *result << create_set_variable_constant(env, state_var, pause_encoding)
	  << create_jump_statement(env, rs->get_exit_label())
	  << create_label_location_statement(env, resume_label);

  return result;
}

/**
  resume { body }

  switch (state) {
    case 0: goto L0;
    case 1: goto L1;
    case 2: goto L2;
  }
  body
Exit:
 */
AnnotableObject *
Dismantler<ResumeStatement>::dismantle(ResumeStatement *rs)
{
  Expression *e =
    create_load_variable_expression(env,
				    rs->get_state_var()->get_type()->get_base_type(),
				    rs->get_state_var());
  MultiWayBranchStatement *mwbs =
    create_multi_way_branch_statement(env, e, 0);

  // Attach the case labels

  int index = 0;
  for (Iter<CodeLabelSymbol *> iter = rs->get_resume_point_iterator() ;
       iter.is_valid() ; iter.next())
    mwbs->add_case(index++, iter.current());

  Statement *body = rs->get_body();
  rs->set_body(0);

  StatementList *result = create_statement_list(env);
  *result << mwbs
	  << body
	  << create_label_location_statement(env, rs->get_exit_label());
    
  return result;
}

/**
  Dismantle a nothing statement into an empty sequence of instructions
 */
AnnotableObject *
Dismantler<NothingStatement>::dismantle(NothingStatement *rs)
{
  return create_statement_list(env);
}

// FIXME: dismantle_exec_statement

// FIXME: dismantle_signal_statement
// Include initializers

// FIXME: dismantle_var_statement
// Include initializers

/**
  exit @k

  exit_level_var := k
  goto Exit_label
 */
AnnotableObject *
Dismantler<ExitAtStatement>::dismantle(ExitAtStatement *eas)
{
  // Locate the innermost enclosing Thread

  SuifObject *so = eas->get_parent();
  while (!is_a<ThreadStatement>(so))
    so = so->get_parent();
  suif_assert(so);
  ThreadStatement *ts = to<ThreadStatement>(so);

  VariableSymbol *exit_var = ts->get_exit_var();
 
  StatementList *result = create_statement_list(env);
  *result << create_set_variable_constant(env, exit_var, eas->get_level())
	  << create_jump_statement(env, ts->get_join_label());
					       
  return result;
}

/**
  Fix Esterel ProcedureDefinitions so that they are more SUIF-like
 */
/*
void FixProcDefs::do_procedure_definition(ProcedureDefinition *pd)
{
  SuifEnv *env = pd->get_suif_env();

  EsterelInformationBlock *info_block =
    to<EsterelInformationBlock>(get_global_information_block(env,
		   EsterelInformationBlock::get_class_name()));
  suif_assert(info_block); // Should have been attached by strl2suif

  TypeBuilder *tb = get_type_builder(env);
  suif_assert(tb);
  ProcedureSymbol *ps = pd->get_procedure_symbol();
  suif_assert(ps);

  if ( !(ps->get_type()) ) {
    // Give the procedure definition a type

    list<QualifiedType *> args;
    CProcedureType *cpt =
      tb->get_c_procedure_type(info_block->get_integer_type(), args);
    ps->set_type(cpt);
  }
}
*/

/**
  Build the C interface for each module

  This requires dismantle_parallel_trap_exit to have run
 */
void BuildCInterface::do_procedure_definition(ProcedureDefinition *pd)
{

  ProcedureSymbol *ps = pd->get_procedure_symbol();
  if (is_a<ModuleDefinition>(pd) && !ps->get_type()) {
    ModuleDefinition *md = to<ModuleDefinition>(pd);
    SuifEnv *env = md->get_suif_env();

    suif_assert( is_a<DefinitionBlock>(md->get_parent()) );
    suif_assert( is_a<FileBlock>(md->get_parent()->get_parent()) );
    FileBlock *fb = to<FileBlock>(md->get_parent()->get_parent());

    EsterelInformationBlock *info_block =
      to<EsterelInformationBlock>(get_global_information_block(env,
		   EsterelInformationBlock::get_class_name()));
    suif_assert(info_block); // Should have been attached by strl2suif

    TypeBuilder *tb = get_type_builder(env);
    suif_assert(tb);

    // Set the type of the reaction function to no arguments, return int

    list<QualifiedType *> args;
    CProcedureType *int_noargs_type =
      tb->get_c_procedure_type(info_block->get_integer_type(), args);
    ps->set_type(int_noargs_type);

    // Add a label after the thread statement that is the body 

    // The body should be ThreadStatement created by ParallelTrapExitModule::
    suif_assert( is_a<ThreadStatement>(md->get_body()) );
    ThreadStatement *ts = to<ThreadStatement>(md->get_body());
    md->set_body(0);

    // Start building the int ModuleName_reset() function
    // Put the symbol in the file set block's external symbol table to make
    // it externally visible (i.e., not static)

    ProcedureSymbol *rps = create_procedure_symbol(env, int_noargs_type);
    LString reset_name = ps->get_name() + "_reset";
    env->get_file_set_block()->get_external_symbol_table()->add_symbol(reset_name, rps);

    StatementList *reset_body = create_statement_list(env);
    ProcedureDefinition *reset_proc =
      create_procedure_definition(env, rps, reset_body);
    fb->get_definition_block()->append_procedure_definition(reset_proc);

    // FIXME
    // Add code to reset the outermost thread's state to 0

    StatementList *new_body = create_statement_list(env);

    // Start the reaction function with an emission of the "tick" signal

    SymbolTableObject *sto =
      md->get_signal_symbols()->lookup_lookup_table(tickSignal);

    if (sto) {
      SignalSymbol *tick = to<SignalSymbol>(sto);
      suif_assert(tick->get_presence());
      *new_body << create_set_variable_constant(env, tick->get_presence(), 1);
    }

    *new_body << ts
	      << create_label_location_statement(env, ts->get_join_label());

    // Create input and output functions for each signal

    for ( Iter<SymbolTableObject*> iter =
	    md->get_signal_symbols()->get_symbol_table_object_iterator() ;
	  iter.is_valid() ; iter.next() )
      if (is_a<SignalSymbol>(iter.current())) {
	SignalSymbol *ss = to<SignalSymbol>(iter.current());
	LString direction = ss->get_direction();

	if ( direction == inputDirection ||
	     direction == inputoutputDirection ) {

	  // Add reset code for input and input/output signals in both the
	  // reaction and reset functions

	  *new_body <<
	    create_set_variable_constant(env, ss->get_presence(), 0);
	  *reset_body <<
	    create_set_variable_constant(env, ss->get_presence(), 0);
	}

	// FIXME

	// Add input functions for input signals

	if (direction == outputDirection ||
	    direction == inputoutputDirection) {

	  // An output signal: create a declaration for the output
	  // function, i.e., ModuleName_O_SignalName

	  LString output_name = ps->get_name() + "_O_" + ss->get_name();
	  list<QualifiedType *> args;
	  if (ss->get_type() != info_block->get_q_void_type()) {
	    // A valued signal: give the output function a single argument
	    args.push_back(ss->get_type());
	  }
	  CProcedureType *output_type =
	    tb->get_c_procedure_type(info_block->get_void_type(), args);
	  ProcedureSymbol *ops = create_procedure_symbol(env, output_type);
	  fb->get_symbol_table()->add_symbol(output_name, ops);

	  // Add code to the reaction function that checks signal
	  // presence for output and calls those functions

	  SymbolAddressExpression *sae =
	    create_symbol_address_expression(env,
		     tb->get_pointer_type(output_type), ops);

	  CallStatement *cos =
	    create_call_statement(env, (VariableSymbol*)0, sae);
	  if (ss->get_type() != info_block->get_q_void_type()) {
	    // A valued signal: output the value as its argument
	    cos->append_argument(
		 create_load_variable(env, ss->get_value() ));
	  }
	
	  *new_body << create_if_statement(env,
	     create_load_variable(env, ss->get_presence()),
	     cos);
	}
      }

    // Create a return statement at the end of the body

    *new_body << create_return_statement(env,
  	           create_load_variable_expression(env,
			ts->get_exit_var()->get_type()->get_base_type(),
						   ts->get_exit_var()));
    md->set_body(new_body);    
  }
}

#if 0
void ParallelTrapStatementDismantler::do_procedure_definition(ProcedureDefinition *pd)
{
  if (!is_a<ModuleDefinition>(pd)) return;
  ModuleDefinition *md = to<ModuleDefinition>(pd);

  Statement *st = to<Statement>(md->get_body());

  Context c;
  c.depth = 0;

  walk_statement(st, c);
}

void ParallelTrapStatementDismantler::walk_statement(Statement *st,
						     const Context &c)
{
  printf("%*s walking a %s\n", c.depth, "", st->getClassName().c_str());
  for ( Iter<Statement> iter = collect_instance_objects<Statement>(st) ;
	iter.is_valid() ; iter.next() ) {
    Context nc;
    nc.depth = c.depth + 1;
    walk_statement(&iter.current(), nc);
  }
}
#endif

